/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2005-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: target_kernel_impl.c 301 2015-01-07 04:57:01Z ertl-ishikawa $
 */

/*
 * ターゲット依存モジュール（BlueNinja用）
 */
#include "kernel_impl.h"
#include <sil.h>
#include "tz1000.h"

/*
 *  東芝製ドライバを使用するためのI/F
 */
uint32_t SystemCoreClock = 12000000;

/*
 *  ターゲット依存の初期化
 */
void
target_initialize(void){

    volatile uint32_t i;

	/*
	 * コア依存の初期化
	 */
	core_initialize();

    sil_orw((void *)CPACR_BASE, ((3UL << 10*2) | (3UL << 11*2)));

    /*
     *  wait 2sec
     */
    for (i = 0; i < 2000U; i++) {
        sil_dly_nse(1000000);
    }

    /* PMU Initialize */
    Driver_PMU.Initialize(NULL);
    Driver_PMU.SelectClockSource(PMU_CSM_MAIN, PMU_CLOCK_SOURCE_PLL);
    Driver_PMU.SetPrescaler(PMU_CD_MPIER, 1);
    Driver_PMU.SetPrescaler(PMU_CD_PPIER0, 2);
    Driver_PMU.SetPrescaler(PMU_CD_PPIER1, 4);
    Driver_PMU.SetPrescaler(PMU_CD_PPIER2, 4);
    /*
     * GPIO
     *  GPIO1 ---> PW switch
     *  GPIO2 ---> reset
     *  GPIO3 ---> Hold
     *  GPIO4 ---> battery
     *  GPIO10 ---> LED RED
     *  GPIO11 ---> LED GREEN
     *
     */
    /* PW switch */
    Driver_PMU.StandbyInputBuffer(PMU_IO_FUNC_GPIO_1, 0);
    Driver_PMU.StandbyInputBuffer(PMU_IO_FUNC_GPIO_4, 0);

    /* GPIO Initialize */
    Driver_GPIO.Initialize();
    Driver_GPIO.PowerControl(ARM_POWER_FULL);
    /* config Input/Output */
    Driver_GPIO.Configure(1,  GPIO_DIRECTION_INPUT_HI_Z, GPIO_EVENT_DISABLE, NULL);
    Driver_GPIO.Configure(3, GPIO_DIRECTION_OUTPUT_2MA, GPIO_EVENT_DISABLE, NULL);
    Driver_GPIO.Configure(4, GPIO_DIRECTION_INPUT_HI_Z, GPIO_EVENT_DISABLE, NULL);
    Driver_GPIO.Configure(11, GPIO_DIRECTION_OUTPUT_2MA, GPIO_EVENT_DISABLE, NULL);

    /* GPIO3 High Hold / LED(G) turn on */
    Driver_GPIO.WritePin(3, 1);
    Driver_GPIO.WritePin(11, 1);

    /* 
     * UART Initialize
     * banner.c(target_fput_log)のためにここで初期化する
     */
    Driver_PMU.SelectClockSource(PMU_CSM_UART1, PMU_CLOCK_SOURCE_OSC12M);
    Driver_PMU.SetPrescaler(PMU_CD_UART1, 1);
    Driver_UART1.Initialize(0, 0);
    Driver_UART1.Configure(115200, 8, ARM_UART_PARITY_NONE, 
            ARM_UART_STOP_BITS_1, ARM_UART_FLOW_CONTROL_NONE);
    Driver_UART1.PowerControl(ARM_POWER_FULL);
} 


/*
 *  ターゲット依存の終了処理
 */
void
target_exit(void)
{
    /* GPIO3 Low Hold / LED(G) turn off */
    Driver_GPIO.WritePin(11, 0);
    Driver_GPIO.WritePin(3, 0);

	/*
	 *  コア依存の終了処理
	 */
	core_terminate();

	/*
	 *  開発環境依存の終了処理
	 */
	while(1);
}


/*
 *  システムログの低レベル出力のための文字出力
 */
void
target_fput_log(char c)
{
	if (c == '\n') {
        while ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_TXFF) == UARTFR_TXFF);
        sil_wrw_mem(UARTDR(UART1_BASE), (uint32_t)('\r'));
	}
    while ((sil_rew_mem(UARTFR(UART1_BASE)) & UARTFR_TXFF) == UARTFR_TXFF);
    sil_wrw_mem(UARTDR(UART1_BASE), (uint32_t)((uint8_t)(c)));
}

void
check_power_handler(intptr_t exinf)
{
    static uint8_t check_count = 0U;
    static bool_t led_out = true;
    uint32_t check_power_switch;
    uint32_t check_battery;

    /*
     * PW switch の押下状況チェック
     * バッテリ状況チェック
     */
    Driver_GPIO.ReadPin(1, &check_power_switch);
    Driver_GPIO.ReadPin(4, &check_battery);

    if (!(check_power_switch & 0x1U)) {
        check_count++;
    }
    else if (!(check_battery & 0x1U)) {
        check_count++;
    }
    else {
        check_count = 0U;
    }

    if (check_count >= 5U) {
        /*
         * 電源OFF判定5秒以上
         */
        ext_ker();
    }
    /*
     *  LED(G)反転処理 
     */
    else if (led_out) {
        Driver_GPIO.WritePin(11, 0);
        led_out = false;
    }
    else {
        Driver_GPIO.WritePin(11, 1);
        led_out = true;
    }
}

